#!/bin/bash

# Guard representation:
# guards are boxes generated with 2 arguments:
# Argument 1: Distance from (0,0) to the guard point that is closest to the flowpipe
# Argument 2: Angle in degrees from (0,0) to the closest point (/vertex)
# Argument 3: If given, box is inside the circle

# 2 guards: one inside circle one outside:
GUARDS=( "1.01 -30" "0.79 -40 -inside" )
# 1 guard for plots:
#GUARDS=( "0.625 -35 -inside" )

# Refinement modes to be used
# All:
MODES=( "CGS" "CAS" "PGS" "PAS" "CGD" "CAD" "PGD" "PAD" "CGB" "CAB" "PGB" "PAB" )
# Only 1 mode for plots:
#MODES=( "PAS" )

# Global commands for all modes (plots, number of max iterations, etc.)
GLOBALCOMMANDS=""
#GLOBALCOMMANDS="--stc-direction-refinement-plot-steps -v D7"


LATEX_FILE="table_circle.tex"
FILE="circle_benchmark"
PLOTDIR="plots"
OUTPUTDIR="output"
TEMPLATE_LINE="ID & ALGO & CALLS & NUMDIRS & TIME & RESULT \\\\"

rm -fv ${LATEX_FILE}
rm -rf ${FILE}*.txt
rm -rf ${PLOTDIR}/*
rm -rf ${OUTPUTDIR}

if [ -d plots ]; then
	rm -rfv plots/*
else
	mkdir plots
fi

if [ -d output ]; then
	rm -rfv output/*
else
	mkdir output
fi

rm -f $LATEX_FILE

function table_header {
	echo "\\begin{longtable}{|c|c|c|c|c|c|}" >> $LATEX_FILE
	echo "\\hline" >> $LATEX_FILE
	echo "ID & Algorithm & \\# IR & \\# d & Runtime & Res. \\\\" >> $LATEX_FILE
	echo "\\hline" >> $LATEX_FILE
}

function table_footer {
	echo "\\end{longtable}" >> $LATEX_FILE
}

function write_results {
    	#Result: Separtion or not?
	local RES=`cat $OUTPUTFILE | grep -e "Forbidden states are" | rev | cut -d ' ' -f2 | rev`

	local RESULT=""

	if [ "$RES" == "not" ]; then
		RESULT="SEP"
	else
		if [ "$RES" == "are" ]; then
			RESULT="INT"
		else
			RESULT="?"
			echo "ERROR: Could not parse output file. Instance ${ID} mode ${MODE}, Line in question: $RES" >> error.log
		fi
	fi
	LINE=`echo $LINE | sed "s/RESULT/${RESULT}/g"`

        # Runtime
	local RUNTIME=`cat $OUTPUTFILE | grep "Refining flowpipe intersection intervals done after" | rev | cut -d ' ' -f1 | rev`

	local RUNTIME=`echo ${RUNTIME} | sed "s/s//g"`

	local RUNTIME=`bc <<< "scale=3; ( ${RUNTIME} / 1 )"`

	RUNTIME=`echo ${RUNTIME}s | sed ':a;N;$!ba;s/\\n/ /g'`
	LINE=`echo $LINE | sed "s/TIME/${RUNTIME}/g"`
	
	REFINEMENTS=`cat $OUTPUTFILE | grep -e "Step of the refinement algorithm done after" | wc -l`
	if [ $REFINEMENTS == "0" ]; then
	    REFINEMENTS=`cat $OUTPUTFILE | grep -e "Global NLOpt optimization done after" | wc -l`
	fi
	LINE=`echo $LINE | sed "s/CALLS/${REFINEMENTS}/g"`

	NUMDIRS=`cat $OUTPUTFILE | grep -e "Computing flowpipe support value" | wc -l`

	if [ "$NUMDIRS" == 0 ]; then
	   NUMDIRS=`cat $OUTPUTFILE | grep -e "Computing difference support value... \$" | wc -l` 
	fi 

	LINE=`echo $LINE | sed "s/NUMDIRS/${NUMDIRS}/g"`

}

function read_mode_commands {
        
    MODECOMMANDS="$GLOBALCOMMANDS"

    # Check if separation is to be used at all
    if [ "$MODE" != "NONE" ]; then

        # We do need additional options	for refinement
	MODECOMMANDS="$MODECOMMANDS --stc-direction-refinement"
	
	# Check if any non-default values are to be set:
	# Point intervall enabled?
	POINT=`echo $MODE | grep P`
	if [ "$POINT" != "" ]; then
	    MODECOMMANDS="$MODECOMMANDS --stc-direction-refinement-time-mode point"
	fi
	
	# Global optimization enabled?
	REFMODE=`echo $MODE | grep N`
	if [ "$REFMODE" != "" ]; then
	    MODECOMMANDS="$MODECOMMANDS --stc-direction-refinement-separation-algorithm nlopt" 
	fi
	REFMODE=`echo $MODE | grep G`
        if [ "$REFMODE" != "" ]; then
	    MODECOMMANDS="$MODECOMMANDS --stc-direction-refinement-separation-algorithm gjk"
	    MODECOMMANDS="$MODECOMMANDS --stc-direction-refinement-use-facet-normals no"
	fi
	REFMODE=`echo $MODE | grep A`
        if [ "$REFMODE" != "" ]; then
	    MODECOMMANDS="$MODECOMMANDS --stc-direction-refinement-separation-algorithm approx"
	fi
	
	# Static interval refinement enabled?
	STAT=`echo $MODE | grep S`
	if [ "$STAT" != "" ]; then
	    MODECOMMANDS="$MODECOMMANDS --stc-direction-refinement-interval-refinement-mode stat"
	fi

	# Dynamic interval refinement enabled?
	DYN=`echo $MODE | grep D`
	if [ "$DYN" != "" ]; then	    
	    MODECOMMANDS="$MODECOMMANDS --stc-direction-refinement-interval-refinement-mode dyn"
	fi
	REFMODE=`echo $MODE | grep B`
        if [ "$REFMODE" != "" ]; then
	    MODECOMMANDS="$MODECOMMANDS --stc-direction-refinement-interval-refinement-mode both"
	fi
    fi
}

function write_algo {
    # Write Algorithm information to latex table line
    ALGOSTRING=""

    TMPSTRING=`echo ${MODE} | grep C`
    	if [ "$TMPSTRING" != "" ]; then
	    ALGOSTRING="${ALGOSTRING}CH, "
	fi
    TMPSTRING=`echo ${MODE} | grep P`
	if [ "$TMPSTRING" != "" ]; then
	    ALGOSTRING="${ALGOSTRING}PW, "
	fi
    TMPSTRING=`echo ${MODE} | grep G`
    	if [ "$TMPSTRING" != "" ]; then
	    ALGOSTRING="${ALGOSTRING}GJK"
	fi
    TMPSTRING=`echo ${MODE} | grep A`
	if [ "$TMPSTRING" != "" ]; then
	    ALGOSTRING="${ALGOSTRING}DA"
	fi
     TMPSTRING=`echo ${MODE} | grep S`
        if [ "$TMPSTRING" != "" ]; then
	   ALGOSTRING="${ALGOSTRING}, Fix."
	fi
     TMPSTRING=`echo ${MODE} | grep D`
        if [ "$TMPSTRING" != "" ]; then
	   ALGOSTRING="${ALGOSTRING}, Dyn."
	fi
     TMPSTRING=`echo ${MODE} | grep B`
        if [ "$TMPSTRING" != "" ]; then
	   ALGOSTRING="${ALGOSTRING}, Both"
	fi
     echo "ALGO String: $ALGOSTRING"

     LINE=`echo $LINE | sed "s/ALGO/${ALGOSTRING}/g"`
}

function run_spaceex {
    # Run spaceex with given options
    OUTPUTFILE="${OUTPUTDIR}/screen-${MODE}-ID${ID}.txt"

    echo "Running SpaceEx with mode = $MODE"
    
    read_mode_commands

    echo "MODECOMMANDS = ${MODECOMMANDS}"

    # Run spaceex with the given refinement modes
    local CMD="-m circle.xml --config circle.cfg --output-file ${PLOTDIR}/plot_${MODE}_ID${ID} ${MODECOMMANDS} "

    time ../../spaceex/spaceex $CMD > $OUTPUTFILE

    # Write the results into LINE
    write_results
}

# Run refinement algorithm in every setting on every specified guard
rm -rf output/* *.tex *.log
ID="1"

for guard in "${GUARDS[@]}"
do
table_header

# STEP 0: Prepare .xml 
    rm -f ./circle.xml
    cp circle_with_guard_placeholder.xml circle.xml

# Create the .xml file for the current instance/ guard
    guard_string=`./guard_generator $guard`
    #./guard_generator $guard >> $guard_string
    echo creating model file with guard= ${guard_string} generated from $guard
    sed -i "s/GUARD_PLACE_HOLDER/${guard_string}/g" circle.xml
    sed -i "s/\\\//g" circle.xml
    sed -i "s/\\/amp/\\\\\&amp/g" circle.xml
    sed -i "s/\\/lt/\\\\\&lt/g" circle.xml
    sed -i "s/\\/gt/\\\\\&gt/g" circle.xml
    sed -i "s/\\\//g" circle.xml


# Run spaceex on current iteration in every refinement mode and save results in LINE
    for MODE in "${MODES[@]}"; do
	LINE=$(echo $TEMPLATE_LINE | sed s/MODE/${MODE}/g)
	LINE=`echo $LINE | sed "s/ID/${ID}/g"`
	write_algo
	run_spaceex

	echo $LINE
	echo $LINE >> $LATEX_FILE
	echo "\\hline" >> $LATEX_FILE
	ID=$(($ID + 1))

    done 

    #echo "\\hline" >> $LATEX_FILE

table_footer

done
