#!/bin/bash

###############################################################################

# This script calls SpaceEx' direction refinement algorithms on the sphere
# benchmark. The generation of the benchmark was added to the functionality of
# SpaceEx. The command-line parameters concerning this sphere benchmark
# generation start with --stc-direction-refinement-benchmark...
#
# The script can be called with the following parameters:
#
# $ ./benchmark.sh <ALGO> <MODE> <GUARD> <MAX_FUNC> <SLAB_MODE>
#
# <ALGO>:
#		NLOPT	- use NLOpt to determine separation with the guard
#		IRON	- use the modified GJK to determine separation
#		DIAMOND - use directed approximation to determine separation
#
# <MODE>:
#		rnd	- Generate a random direction in which the guard lies
#		fixed	- Generate the vector (1,...,1) as direction for the
#			  guard
#
# <GUARD>:
#		box	- Generate a guard shaped like a box with side length 1
#		triangle - Generate a guard shaped like a triangle where the
#			   slope is perpendicular to the sphere
#		vertex	- Generate a guard which is a single vertex instead of
#			  a polytope
#
# <MAX_FUNC>:
#		cheby	- Compute the chebyshev center to be calculated for
#			  simplex points on the difference polyhedra
#		max	- Use the maximum distance for simplex points
#		maxnlopt - Use NLOpt to determine the maximum distance for
#			   simplex points
#
# <SLAB_MODE>:
#		slab	- Use facet slabs during the calculation
#		slice	- Use facet slices during the calculation

# The variable K defines the distances of the guard to the sphere to be used.
# Distances may also be negative. Note, however, that negative distances mean
# that the guard is INSIDE of the sphere (except the negative distance is too
# small, leading to the guard being outide of the sphere in the opposite
# direction.
K="0.01 0.1 0.5 1"

# The variable D defines the dimensions to be analyzed. Reduce the maximum
# dimension to decrease the time needed for SpaceEx to analyze the problem.
D="2 3 4 5"

###############################################################################

LATEX_FILE="table_sphere"

FILE="sphere_benchmark"
PLOTDIR="plots"
OUTPUTDIR="output"


# Limit to 15 minutes = 900 sec
#SPACEEX_TIMEOUT=900s
SPACEEX_TIMEOUT=500s

rm -fv ${LATEX_FILE}
rm -rf ${FILE}*.txt
rm -rf ${PLOTDIR}/*
rm -rf ${OUTPUTDIR}

if [ ! -d plots ]; then
	mkdir plots
fi

mkdir ${OUTPUTDIR}

ALGO_STRING=$1

case $ALGO_STRING in 
	 "NLOPT")
	      ALGO="NLOPT"
	      ;;
	 "IRON")
	      ALGO="IRON"
	      ;;
	 "GOLD")
	      ALGO="GOLD"
	      ;;
	 "DIAMOND")
	      ALGO="DIAMOND"
	      ;;
	 *)
	      ALGO=
esac

MODE_STRING=$2
GUARD_TYPE=$3

MAXIMIZATION=$4
SLAB_MODE=$5

# Auxiliary variables that can alter how the script behaves
if [ "$#" -gt 7 -a "$#" -lt 8 ]; then
	echo "Error: Number of command line paramter mismatch. Expected number < 7 or 8!"
	exit 1
fi

if [ -n "$6" ]; then
	BENCHMARK_NUMBER="$6"
fi

if [ "$#" -eq 8 ]; then
	BENCHMARK_NUMBER="$6"
	INPUT_DIM="$7"
	FAC_NUM="$8"
fi

MODE_TYPE=""

if [ -n "${MODE_STRING}" ]; then
	
	case "${MODE_STRING}" in
	"rnd")
		# Random mode
		MODE_CMD="--stc-direction-refinement-benchmark-random-direction"
		;;
	"fixed")
		# Use fixed (1,...,1) direction
		MODE_CMD=""
		;;
	*)
		echo "ERROR: Unknown mode ${MODE_STRING}"
		exit 1
		;;
	esac

	MODE_TYPE="${MODE_STRING}"
fi

if [ -n "$GUARD_TYPE" ]; then

	case "$GUARD_TYPE" in
	"box")
		GUARD_CMD="--stc-direction-refinement-guard-type box"
		;;
	"triangle")
		GUARD_CMD="--stc-direction-refinement-guard-type triangle"
		;;
	"vertex")
		GUARD_CMD="--stc-direction-refinement-guard-type vertex"
		;;
	*)
		echo "ERROR: Unknown guard type ${GUARD_TYPE}"
		exit 1
		;;
	esac

	MODE_TYPE="${MODE_TYPE}-${GUARD_TYPE}"
fi

LATEX_FILE="${LATEX_FILE}-${MODE_TYPE}-${MAXIMIZATION}-${SLAB_MODE}"

if [ -n "$BENCHMARK_NUMBER" ]; then
	LATEX_FILE="${LATEX_FILE}_${BENCHMARK_NUMBER}"
fi

LATEX_FILE="${LATEX_FILE}-${ALGO}.tex"

function table_header {

	if [ -n "${MODE_STRING}" ]; then
		echo "\\section{Mode: ${MODE_STRING} --- Guard Shape: $GUARD_TYPE -- Max. Function: $MAXIMIZATION -- Facet mode: $SLAB_MODE}" >> $LATEX_FILE
	fi

	echo -n "\\begin{longtable}{|c|c|c|c|c|c|c|c|" >> $LATEX_FILE
	if [ "${ALGO}" == "NLOPT" ]; then
		echo -n "c|c|" >> $LATEX_FILE
	fi
	
#	if [ "${MODE_STRING}" == "rnd" ]; then
#		echo -n "c|" >> $LATEX_FILE
#	fi
	echo "}" >> $LATEX_FILE
	echo "\\hline" >> $LATEX_FILE
	echo -n "ID & Dim & \\#Fac & \$k\$ & Exp & Result & \\#Iterations & Time " >> $LATEX_FILE
	if [ "${ALGO}" == "NLOPT" ]; then
		echo -n "& Opt. Val & NLOPT " >> $LATEX_FILE
	fi
	
#	if [ "${MODE_STRING}" == "rnd" ]; then
#		echo -n "& Vector " >> $LATEX_FILE
#	fi
	echo "\\\\" >> $LATEX_FILE
	echo "\\hline" >> $LATEX_FILE
}

function table_footer {
	echo "\\end{longtable}" >> $LATEX_FILE
}

function prepare_spaceex {

	local INSTANCE="$1"
	local DIMENSION="$2"
	local FACETS="$3"
	local DISTANCE="$4"
	local MODE="$5"

	local MODEL="${FILE}.xml"
	local CFG="${FILE}.cfg"
	local OUTPUT="${PLOTDIR}/${FILE}-plot-${INSTANCE}-${MODE_TYPE}-${MAXIMIZATION}-${SLAB_MODE}-${MODE}-${USE_NORMALS}.txt"
	local SYSTEM="\"aut\""
	
	# The following variables are mockup and not actually needed for this sample
	local SCENARIO="\"stc\""
	local DIRECTIONS="\"box\""
	local FP_TOL="0.01"
	local TIME_HORIZON="10"
	local ITER_MAX="500"
	
	# Set the output variables in case of plotting
	local OUTPUT_VARS="\"sphere_var_0, sphere_var_1\""
	local OUTPUT_FORMAT="\"GEN\""
	
	# Relative and absolute error values
	local REL_ERR="1.0e-6"
	local ABS_ERR="1.0e-9"
	
	CMD=( -m ${MODEL} )
	CMD=( "${CMD[@]}" --config ${CFG} )
	CMD=( "${CMD[@]}" --output-file ${OUTPUT} )
	CMD=( "${CMD[@]}" --stc-direction-refinement --stc-direction-refinement-benchmark )

	CMD=( "${CMD[@]}" --output-variables "${OUTPUT_VARS}" )
	CMD=( "${CMD[@]}" --rel-err "${REL_ERR}" )
	CMD=( "${CMD[@]}" --abs-err "${ABS_ERR}" )
	
	CMD=( "${CMD[@]}" --stc-direction-refinement-benchmark-dimension ${DIMENSION} )
	CMD=( "${CMD[@]}" --stc-direction-refinement-benchmark-facets ${FACETS} )
	CMD=( "${CMD[@]}" --stc-direction-refinement-benchmark-guard-distance \"${DISTANCE}\" )

	if [ -n "$USE_FACET_NORMALS" ]; then
		CMD=( "${CMD[@]}" ${USE_FACET_NORMALS} )
	fi

	if [ -n "$MODE_CMD" ]; then
		CMD=( "${CMD[@]}" ${MODE_CMD} )
	fi

	if [ -n "$GUARD_CMD" ]; then
		CMD=( "${CMD[@]}" ${GUARD_CMD} )
	fi

	if [ -n "$MAXIMIZATION" ]; then
		local SLAB_ALGO="--stc-direction-refinement-slab-algorithm"
		case "$MAXIMIZATION" in
		"max")
			SLAB_ALGO="${SLAB_ALGO} max"
			;;
		"maxnlopt")
			SLAB_ALGO="${SLAB_ALGO} max-nlopt"
			;;
		"cheby")
			SLAB_ALGO="${SLAB_ALGO} chebyshev"
			;;
		*)
			echo "Unknown slab mode ${MAXIMIZATION}"
			exit 1
		esac

		CMD=( "${CMD[@]}" ${SLAB_ALGO} )
	fi

	if [ -n "$SLAB_MODE" ]; then
		local SLAB="--stc-direction-refinement-facet-mode"
		case "$SLAB_MODE" in
		"slab")
			SLAB="${SLAB} slab"
			;;
		"slice")
			SLAB="${SLAB} slice"
			;;
		*)
			echo "Unknown slab representation ${SLAB_MODE}"
			exit 1
		esac

		CMD=( "${CMD[@]}" ${SLAB} )
	fi

	if [ "$MODE" == "GLOB" ]; then
		CMD=( "${CMD[@]}" --stc-direction-refinement-separation-algorithm nlopt )
	fi

	if [ "$MODE" == "APPROX" ]; then
		CMD=( "${CMD[@]}" --stc-direction-refinement-separation-algorithm approx )
	fi
}


function call_spaceex-gjk {

	local INSTANCE="$1"
	local DIMENSION="$2"
	local FACETS="$3"
	local DISTANCE="$4"

	echo Running SpaceEx GJK for instance ${INSTANCE} ...
	
	prepare_spaceex $INSTANCE $DIMENSION $FACETS $DISTANCE "GJK"

	echo ${CMD[@]}

	local SCREEN_NAME="${FILE}-screen-${INSTANCE}-${MODE_TYPE}-${MAXIMIZATION}-${SLAB_MODE}-GJK-${USE_NORMALS}.txt"

	trap 'kill -INT -$pid; echo -e "\nKilled!"' INT
	timeout $SPACEEX_TIMEOUT ../../spaceex/spaceex "${CMD[@]}" > ${OUTPUTDIR}/${SCREEN_NAME} &
	pid=$!
	wait $pid
}

function call_spaceex-approx {

	local INSTANCE="$1"
	local DIMENSION="$2"
	local FACETS="$3"
	local DISTANCE="$4"

	echo Running SpaceEx Approx for instance ${INSTANCE} ...
	
	prepare_spaceex $INSTANCE $DIMENSION $FACETS $DISTANCE "APPROX"

	echo ${CMD[@]}

	local SCREEN_NAME="${FILE}-screen-${INSTANCE}-${MODE_TYPE}-${MAXIMIZATION}-${SLAB_MODE}-approx-${USE_NORMALS}.txt"

	trap 'kill -INT -$pid; echo -e "\nKilled!"' INT
	timeout $SPACEEX_TIMEOUT ../../spaceex/spaceex "${CMD[@]}" > ${OUTPUTDIR}/${SCREEN_NAME} &
	pid=$!
	wait $pid
}

function call_spaceex-glob {

	local INSTANCE="$1"
	local DIMENSION="$2"
	local FACETS="$3"
	local DISTANCE="$4"

	echo Running SpaceEx GLOB for instance ${INSTANCE} ...

	prepare_spaceex $INSTANCE $DIMENSION $FACETS $DISTANCE "GLOB"

	if [ -n "$MODE_STRING" ]; then

		local SCREEN_NAME_GJK="${FILE}-screen-${INSTANCE}-${MODE_TYPE}-${MAXIMIZATION}-${SLAB_MODE}-GJK-${USE_NORMALS}.txt"
		
#		if [ "$MODE_STRING" == "rnd" ]; then
#			local DIRECTION_LINE=`cat ${OUTPUTDIR}/${SCREEN_NAME_GJK} | grep -e ".*Guard direction:" | rev | cut -d ' ' -f1 | rev`
#			NO_BRACKETS=`echo ${DIRECTION_LINE} | sed "s/\(\[\|\]\)//g"`
#			CMD=( "${CMD[@]}" --stc-direction-refinement-benchmark-direction \"${NO_BRACKETS}\" )
#		fi
	fi

	local SCREEN_NAME="${FILE}-screen-${INSTANCE}-${MODE_TYPE}-${MAXIMIZATION}-${SLAB_MODE}-GLOB-${USE_NORMALS}.txt"
	
	echo ${CMD[@]}

	trap 'kill -INT -$pid; echo -e "\nKilled!"' INT
	timeout $SPACEEX_TIMEOUT sspaceex "${CMD[@]}" > ${OUTPUTDIR}/${SCREEN_NAME} &
	pid=$!
	wait $pid
}

function obtain_results {

	local INSTANCE="$1"
	local SCREEN="$2"
	
	if [ ! -f $SCREEN ]; then
		echo "ERROR: Screen ($SCREEN) for instance $INSTANCE not found!"
		exit 1
	fi
	
	local last_line=`cat $SCREEN | grep -e "^Separation has"`
	
	if [[ ! $last_line == Separation* ]]; then
		echo "The result of SpaceEx was unexpected. Instance $INSTANCE." >> error.log
	fi

	if [[ $last_line == *NOT* ]]; then
		INT_RESULT="INT"
	else
		INT_RESULT="SEP"
	fi

	# Runtime
	RUNTIME=`cat $SCREEN | grep "Running Refinement algorithm done" | rev | cut -d ' ' -f1 | rev | sed 's/s//'`

	# GJK Iterations
	ITER_RES=`cat $SCREEN | grep -e "^Refinement algorithm done after" | sed "s/.*after//" | sed "s/iterations.$//" | tr -d ' '`

	NLOPT_ITER=`cat $SCREEN | grep -e "NLOpt iterations: " | rev | cut -d ' ' -f1 | rev`
}

function parse_result {
	local INSTANCE="$1"
	
	local  SCREEN_GJK="${OUTPUTDIR}/${FILE}-screen-${INSTANCE}-${MODE_TYPE}-${MAXIMIZATION}-${SLAB_MODE}-GJK-${USE_NORMALS}.txt"
	local SCREEN_GLOB="${OUTPUTDIR}/${FILE}-screen-${INSTANCE}-${MODE_TYPE}-${MAXIMIZATION}-${SLAB_MODE}-GLOB-${USE_NORMALS}.txt"
	local SCREEN_APPR="${OUTPUTDIR}/${FILE}-screen-${INSTANCE}-${MODE_TYPE}-${MAXIMIZATION}-${SLAB_MODE}-approx-${USE_NORMALS}.txt"

	case $ALGO in
	  "IRON")
	  	SCR=$SCREEN_GJK
	    ;;
	  "GOLD")
		SCR=$SCREEN_GJK
	    ;;
	  "DIAMOND")
		SCR=$SCREEN_APPR
	    ;;
	  "NLOPT")
		SCR=$SCREEN_GLOB
	    ;;
	  *)
	    echo "Error. Algo $ALGO not found."
	    exit 1
	esac
	
	EXPECTED=`cat $SCR | grep -e "^Separation expected" | sed "s/^Separation expected: //" | tr -d ' '`

	if [ "$EXPECTED" == "YES" ]; then
		EXP="SEP"
	else
		if [ "$EXPECTED" == "NO" ]; then
			EXP="INT"
		else
			echo "FATAL ERROR, UNKNOWN EXPECTED RESULT! Iteration $INSTANCE" >> error.log
		fi
	fi

	case $ALGO in
	  "IRON")
	    obtain_results $INSTANCE $SCREEN_GJK
	    INT_RESULT_FINAL="$INT_RESULT"
	    RUNTIME_FINAL=`bc <<< "scale=4; ( $RUNTIME / 1 )"`s
	    ITER_FINAL="$ITER_RES"
	    ;;
	  "GOLD")
	    obtain_results $INSTANCE $SCREEN_GJK
	    INT_RESULT_FINAL="$INT_RESULT"
	    RUNTIME_FINAL=`bc <<< "scale=4; ( $RUNTIME / 1 )"`s
	    ITER_FINAL="$ITER_RES"
	    ;;
	  "DIAMOND")
	    obtain_results $INSTANCE $SCREEN_APPR
	    INT_RESULT_FINAL="$INT_RESULT"
	    RUNTIME_FINAL=`bc <<< "scale=4; ( $RUNTIME / 1 )"`s
	    ITER_FINAL="$ITER_RES"
	    ;;
	  "NLOPT")
	    obtain_results $INSTANCE $SCREEN_GLOB
	    INT_RESULT_FINAL="$INT_RESULT"
	    RUNTIME_FINAL=`bc <<< "scale=4; ( $RUNTIME / 1 )"`s
	    local NLOPT_RUN=`cat $SCREEN_GLOB | grep -e "NLOpt result:"`
	    
	    if [ -n "$NLOPT_RUN" ]; then
		    NLOPT_VALUE=`echo $NLOPT_RUN | rev | cut -d ' ' -f1 | rev`
	    else
		    NLOPT_VALUE="---"
	    fi
	    
	    local NLOPT_EXIT=`cat $SCREEN_GLOB | grep -e "WARNING:.*NLOpt"`

	    if [ -n "$NLOPT_EXIT" ]; then
		    NLOPT_CODE="WARN"
		    if [[ $NLOPT_EXIT =~ .*optimization.step.*$ ]]; then
			    NLOPT_CODE="OVTL"
		    fi
		    if [[ $NLOPT_EXIT =~ .*exceeded.*$ ]]; then
			    NLOPT_CODE="TO"
		    fi
		    if [[ $NLOPT_EXIT =~ .*stop.value.*$ ]]; then
			    NLOPT_CODE="STOP"
		    fi
	    else
		    NLOPT_CODE="OK"
	    fi
	    ;;
	  *)
	    echo "Error. Algo $ALGO not found."
	    exit 1
	esac
}

function size {
	local NUM=0
	
	for i in $1; do
		NUM=$(( $NUM + 1 ))
	done
	
	SIZE="$NUM"
}

if [ -n "$INPUT_DIM" ]; then
	D="$INPUT_DIM"
fi

if [ -n "$FAC_NUM" ]; then
	MAX_FAC=1      # Do not change this number!
else
	MAX_FAC=2
fi



ID=0

size "$D"
D_SIZE=$SIZE

size "$D"
D_SIZE=$SIZE

#size "$F"
F_SIZE=$MAX_FAC
if [ ! -n "$FAC_NUM" ]; then
	F_SIZE=1
fi

size "$K"
K_SIZE=$SIZE

EXPECTED_INSTANCES=$(( $D_SIZE * $F_SIZE * $K_SIZE ))
EXPECTED_INSTANCES=$(( $EXPECTED_INSTANCES - 1 ))

table_header

for dim in $D; do
	if [ ! -n "$FAC_NUM" ]; then
		FAC_COUNT=2
	else
		FAC_COUNT=1
	fi
	
	while [ $FAC_COUNT -le $MAX_FAC ]; do
		
		if [ -n "$FAC_NUM" ]; then
			fac="$FAC_NUM"
		else
			fac=`bc <<< "( ($dim * 2)^$FAC_COUNT )"`
		fi
	
		for kay in $K; do
			
			echo "Iteration $ID of $EXPECTED_INSTANCES"

			USE_NORMALS="no"
			case $ALGO in
			      "IRON")
			      	USE_FACET_NORMALS="--stc-direction-refinement-use-facet-normals no"
				call_spaceex-gjk $ID $dim $fac $kay
				;;
			      "NLOPT")
				call_spaceex-glob $ID $dim $fac $kay
				;;
			      "GOLD")
			      	USE_FACET_NORMALS="--stc-direction-refinement-use-facet-normals yes"
			      	USE_NORMALS="yes"
			      	call_spaceex-gjk $ID $dim $fac $kay
				;;
			      "DIAMOND")
				call_spaceex-approx $ID $dim $fac $kay
				;;
				
			esac
			
			parse_result $ID

			LINE="$ID & $dim & $fac & $kay & $EXP & $INT_RESULT_FINAL & $ITER_FINAL & $RUNTIME_FINAL"
			
			if [ "$ALGO" == "NLOPT" ]; then
			    LINE="${LINE} & $NLOPT_VALUE & $NLOPT_CODE"
			fi
			
			#LINE="$ID & $dim & $fac & $kay & $EXP & $INT_RESULT_GJK & $INT_RESULT_APPR & $INT_RESULT_NLOPT & $GJK_ITER & $APPR_ITER & $NLOPT_ITER & $RUNTIME_GJK & $RUNTIME_APPR & $RUNTIME_NLOPT & $NLOPT_VALUE & $NLOPT_CODE"

#			if [ "$MODE_STRING" == "rnd" ]; then
#				if [ ! "$GUARD_TYPE" == "box" ]; then
#					LINE="${LINE} & $NO_BRACKETS"
#				fi
#			fi

			LINE="${LINE} \\\\"

			echo $LINE >> $LATEX_FILE
			
			ID=$(( $ID + 1 ))

		done
		echo "\\hline" >> $LATEX_FILE
	
	FAC_COUNT=$(( $FAC_COUNT + 1 ))
	
	done
	echo "\\hline" >> $LATEX_FILE
done

table_footer
